#!/usr/bin/env python3

"""
File to handle the configurations.
"""

import os
import logging
from flask import current_app
import requests

logger = logging.getLogger(__name__)


class Config(object):
    """
    Helper class to hold the configurations and is meant for internal use
    by module only. This is a loose wrapper for flasks application config
    (alias: `app.config`).
    """

    @property
    def get_auth_manager(self):
        auth_manager = current_app.extensions.get("cognito-flask-auth")
        if not auth_manager:
            msg = "Flask Cognito Auth extention is not registerd with Flask application."
            logger.info(msg)
            raise RuntimeError(msg)

        return auth_manager

    def get_config_value(self, key, error_message, is_key_required, is_value_required):
        if key not in current_app.config and is_key_required:
            raise RuntimeError(error_message)

        value = None
        if key in current_app.config:
            value = current_app.config[key]

        if is_value_required and not value:
            raise RuntimeError(error_message)
        return value

    @property
    def client_id(self):
        error_message = 'COGNITO_CLIENT_ID must be set to validate the audience claim.'
        client_id = self.get_config_value(key="COGNITO_CLIENT_ID",
                                          error_message=error_message,
                                          is_key_required=True,
                                          is_value_required=True)
        return client_id

    @property
    def client_secret(self):
        error_message = 'COGNITO_CLIENT_SECRET must be set to validate the audience claim.'
        client_secret = self.get_config_value(key="COGNITO_CLIENT_SECRET",
                                              error_message=error_message,
                                              is_key_required=True,
                                              is_value_required=True)
        return client_secret

    @property
    def user_pool_id(self):
        error_message = 'COGNITO_USER_POOL_ID must be specified to locate the auth url.'
        pool_id = self.get_config_value(key="COGNITO_USER_POOL_ID",
                                        error_message=error_message,
                                        is_key_required=True,
                                        is_value_required=True)
        return pool_id

    @property
    def domain(self):
        error_message = 'COGNITO_DOMAIN must be set to validate the token redirect to create endpoint url.'
        domain = self.get_config_value(key="COGNITO_DOMAIN",
                                       error_message=error_message,
                                       is_key_required=True,
                                       is_value_required=True)
        if not domain.startswith("https://"):
            domain = f"https://{domain}"
        return domain

    @property
    def region(self):
        error_message = 'COGNITO_REGION must be specified.'
        region = self.get_config_value(key="COGNITO_REGION",
                                       error_message=error_message,
                                       is_key_required=True,
                                       is_value_required=True)
        return region

    @property
    def redirect_uri(self):
        error_message = 'COGNITO_REDIRECT_URI must be set to obtain callback url.'
        uri = self.get_config_value(key="COGNITO_REDIRECT_URI",
                                    error_message=error_message,
                                    is_key_required=True,
                                    is_value_required=True)
        return uri

    @property
    def redirect_error_uri(self):
        uri = self.get_config_value(key="ERROR_REDIRECT_URI",
                                    error_message=None,
                                    is_key_required=False,
                                    is_value_required=False)
        return uri

    @property
    def signout_uri(self):
        error_message = 'COGNITO_SIGNOUT_URI must be set for logout callback.'
        uri = self.get_config_value(key="COGNITO_SIGNOUT_URI",
                                    error_message=error_message,
                                    is_key_required=True,
                                    is_value_required=True)
        return uri

    @property
    def exempt_methods(self):
        methods = self.get_config_value(key="EXEMPT_METHODS",
                                        error_message=None,
                                        is_key_required=False,
                                        is_value_required=False)
        return methods if methods else ["OPTIONS"]

    @property
    def issuer(self):
        return f"https://cognito-idp.{self.region}.amazonaws.com/{self.user_pool_id}"

    @property
    def public_key_uri(self):
        return f"{self.issuer}/.well-known/jwks.json"

    @property
    def jwt_code_exchange_uri(self):
        return f"{self.domain}/oauth2/token"

    @property
    def jwt_cognito_key(self):
        auth_manager = self.get_auth_manager
        # load and cache cognito JSON Web Key (JWK)
        jwt_key = None
        if not auth_manager.jwt_key:
            jwt_key = requests.get(self.public_key_uri).json()["keys"]
        else:
            jwt_key = auth_manager.jwt_key
        return jwt_key

    @property
    def state(self):
        csrf_state = self.get_config_value(key="COGNITO_STATE",
                                           error_message=None,
                                           is_key_required=False,
                                           is_value_required=False)
        return csrf_state

    @property
    def login_uri(self):
        csrf_state = self.state
        state_val = ""
        if csrf_state:
            state_val = f"&state={csrf_state}"
        return (f"{self.domain}/authorize?client_id={self.client_id}"
                f"&response_type=code{state_val}"
                f"&redirect_uri={self.redirect_uri}")

    @property
    def logout_uri(self):
        return (f"{self.domain}/logout?response_type=code"
                f"&client_id={self.client_id}&logout_uri={self.signout_uri}")



#!/usr/bin/env python3

"""
File to initalize the AWS Cognito authentor manager.
"""


class CognitoAuthManager(object):
    """
    An object used to hold Flask AWS Cognito settings and callback functions
    for the authentication with AWS Cognito using OAuth2 JWT token.
    Instances of :class:`CognitoAuthManager` are *not* bound to specific apps, so
    you can create one in the main body of your code and then bind it
    to your application.
    Lazy initalization is supported for configuring the application.
    """

    def __init__(self, app=None):
        """
        Create the CognitoAuthManager instance. You can either pass a flask
        application in directly to register the extension with the flask app,
        or call init_app (lazy initalization) after creating the object
        (in a factory pattern).
        :param app: A flask application
        """
        if app is not None:
            self.init(app)
        self.jwt_key = None

    def init(self, app):
        """
        Register this extension with the flask app.
        :param app: A flask application
        """
        # Save this so we can use it later in the extension
        if not hasattr(app, 'extensions'):   # pragma: no cover
            app.extensions = {}
        app.extensions['cognito-flask-auth'] = self



#!/usr/bin/env python

"""
Prerequesites -
  Python Packages:
    * setuptools
    * GitPython
  System Packages:
    * make
    * Python 3
Commands: python setup.py [bdist_wheel / [sdist [--format=[gztar][,zip][,tar]]]
Ex:
  * python setup.py bdist_wheel
  * python setup.py sdist
  * python setup.py sdist --format=gztar
  * python setup.py sdist --format=zip
  * python setup.py sdist --format=tar
  * python setup.py sdist --format=gztar,tar,zip
  * python setup.py sdist --format=gztar,zip
  * python setup.py bdist_wheel sdist --format=gztar,zip,tar
"""

"""
distutils/setuptools install script.
"""


from setuptools import setup, find_packages
import traceback
import shutil
import re
import os
__NAME__ = "flask-cognito-auth"

ROOT = os.path.dirname(os.path.abspath(__file__))
VERSION_FILE = os.path.join(ROOT, __NAME__.replace("-", "_"), ".version")
VERSION_RE = re.compile(r'''__version__ = ['"]([0-9.]+)['"]''')

base = [
    'Flask-WTF',                 # Python flask application
    "requests",                  # Python HTTP for Humans.
    "python_jose"                # JOSE implementation in Python
]

setups = []

ir = (base)
requires = ir


def delete(path):
    if os.path.exists(path=path):
        try:
            if os.path.isfile(path=path):
                os.remove(path=path)
            else:
                shutil.rmtree(path=path)
        except:
            pass


def write_version(version, sha, filename):
    text = "__version__ = '{0}'\n__REVESION__ = '{1}'".format(version, sha)
    with open(file=filename, mode='w') as file:
        file.write(text)


def get_version(filename):
    version = "1.0.0"  # Adding default version

    # This block is for reading the version from Flask Cognito Auth distribution
    if os.path.exists(path=filename):
        contents = None
        with open(file=filename, mode="r") as file:
            contents = file.read()
            version = VERSION_RE.search(contents).group(1)
            return version

    # If file not found. Then may be local or want to get the version
    version_python_file = os.path.join(ROOT, "version.py")
    if os.path.exists(path=version_python_file):
        import version as ver
        version = ver.version

        sha = ""
        try:
            import git
            repo = git.Repo(path=".", search_parent_directories=True)
            sha = repo.head.commit.hexsha
            sha = repo.git.rev_parse(sha, short=6)
        except ImportError:
            print(f"Import error on git, can be ignored for build")
            pass
        except Exception as exception:
            print(str(exception))
            traceback.print_tb(exception.__traceback__)
            pass
        write_version(version=version, sha=sha, filename=filename)
    return version


def replace_string_in_file(old_string, new_string, input, output=None):
    # Safely read the input filename using 'with'
    with open(file=input, mode="r") as f:
        newText = f.read().replace(old_string, new_string)

    if not output or not len(output):
        output = input

    # Safely write the changed content, if found in the file
    with open(file=output, mode="w") as f:
        f.write(newText)


with open("README.md", "r") as f:
    long_description = f.read()


def do_setup():
    setup(
        name=__NAME__,
        version=get_version(filename=VERSION_FILE),
        description='Flask Cognito Authentication',
        long_description=long_description,
        long_description_content_type="text/markdown",
        keywords=['flask', 'amazon cognito', 'json web token',
                  'authentication', 'autorization', 'oauth2', 'saml'],
        author='Ankit Shrivastava',
        url='https://github.com/shrivastava-v-ankit/flask-cognito-auth',
        packages=find_packages(include=[__NAME__.replace("-", "_")]),
        include_package_data=True,
        setup_requires=setups,
        install_requires=requires,
        license="MIT",
        python_requires='>=3.4',
        platforms='any',
        project_urls={
            'Source': 'https://github.com/shrivastava-v-ankit/flask-cognito-auth/',
            'Tracker': 'https://github.com/shrivastava-v-ankit/flask-cognito-auth/issues',
        },
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: Web Environment',
            'Framework :: Flask',
            'Intended Audience :: Developers',
            'Natural Language :: English',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Software Development :: Version Control :: Git',
        ],
    )


if __name__ == '__main__':
    import sys
    do_setup()
    if "sdist" in sys.argv or "bdist_wheel" in sys.argv:
        egg_info = os.path.join(ROOT, __NAME__.replace("-", "_") + '.egg-info')
        delete(path=egg_info)
        eggs = os.path.join(ROOT, '.eggs')
        delete(path=eggs)
        delete(path=VERSION_FILE)
        egg_info = os.path.join(ROOT, "build")
        delete(path=egg_info)



#!/usr/bin/env python3

"""
File handle the decorators for AWS Cognito login / logout features.
On successfull login, add "groups" in session object if user is
part of AWS Cognito group. This helps application for authorization.
"""

import logging
import json
import requests
from requests.auth import HTTPBasicAuth
from functools import wraps
from flask import redirect
from flask import request
from jose import jwt
from .config import Config
from flask import session
from flask import url_for


logger = logging.getLogger(__name__)
config = Config()


def login_handler(fn):
    """
    A decorator to redirect users to AWS Cognito login if they aren't already.
    If already logged in user will redirect redirect uri.
    Use this decorator on the login endpoint.
    This handle will not return to handle the respose rather redirect to
    redirect uri.
    """
    @wraps(fn)
    def wrapper(*args, **kwargs):
        aws_cognito_login = config.login_uri

        # https://docs.aws.amazon.com/cognito/latest/developerguide/amazon-cognito-user-pools-using-tokens-with-identity-providers.html
        res = redirect(aws_cognito_login)
        logger.info("Got Cognito Login, redirecting to AWS Cognito for Auth")
        return res
    return wrapper


def callback_handler(fn):
    """
    A decorator to handle redirects from AWS Cognito login and signup. It
    handles and verifies and exchangs the code for tokens.
    This decorator also pushes the basic informations in Flask session.
    Basic informations are:
        * username
        * group (List of Cognito groups if any)
        * id
        * email
        * expires
        * refresh_token
    Use this decorator on the redirect endpoint on your application.
    """
    @wraps(fn)
    def wrapper(*args, **kwargs):
        auth_success = False
        logger.info("Login is successfull from AWS Cognito.")
        logger.info(
            "Authenticating AWS Cognito application / client, with code exchange.")

        csrf_token = config.state
        csrf_state = None

        if csrf_token:
            csrf_state = request.args.get('state')

        code = request.args.get('code')
        request_parameters = {'grant_type': 'authorization_code',
                              'client_id': config.client_id,
                              'code': code,
                              "redirect_uri": config.redirect_uri}
        response = requests.post(config.jwt_code_exchange_uri,
                                 data=request_parameters,
                                 auth=HTTPBasicAuth(config.client_id,
                                                    config.client_secret))

        # the response:
        # http://docs.aws.amazon.com/cognito/latest/developerguide/amazon-cognito-user-pools-using-tokens-with-identity-providers.html
        if response.status_code == requests.codes.ok:
            logger.info("Code exchange is successfull.")
            logger.info("Validating CSRF state exchange of AWS Cognito")

            if csrf_state == csrf_token:
                auth_success = True

                if csrf_token:
                    logger.info(
                        "CSRF state validation successfull. Login is successfull for AWS Cognito")

                logger.info("Decode the access token from response.")
                verify(response.json()["access_token"])
                id_token = verify(
                    response.json()["id_token"], response.json()["access_token"])

                username = None
                email = None
                if "identities" in id_token:
                    logger.info(
                        "Identities are present in authentication token. Will use that as priority.")

                    for identity in id_token['identities']:
                        if 'primary' in identity and identity['primary']:
                            if 'userId' in identity:
                                email = username = identity['userId']

                    if not username:
                        logger.info(
                            "Primary identifier not present. Will look for very first occurance.")
                        if len(id_token['identities']):
                            identity = id_token['identities'][0]
                            if 'userId' in identity:
                                email = username = identity['userId']

                if not username:
                    username = id_token["cognito:username"]
                if not email and 'email' in id_token:
                    email = id_token["email"]

                groups = None
                if "cognito:groups" in id_token:
                    groups = id_token['cognito:groups']

                update_session(username=username,
                               id=id_token["sub"],
                               groups=groups,
                               email=email,
                               expires=id_token["exp"],
                               refresh_token=response.json()["refresh_token"])
        if not auth_success:
            error_uri = config.redirect_error_uri
            if error_uri:
                resp = redirect(url_for(error_uri))
                return resp
            else:
                msg = f"Somthing went wrong during authentication"
                return json.dumps({'Error': msg}), 500
        return fn(*args, **kwargs)
    return wrapper


def update_session(username: str, id, groups, email: str, expires, refresh_token):
    """
    Method to update the Flase Session object with the informations after
    successfull login.
    :param username (str):      AWS Cognito authenticated user.
    :param id (str):            ID of AWS Cognito authenticated user.
    :param groups (list):       List of AWS Cognito groups if authenticated
                                user is subscribed.
    :param email (str):         AWS Cognito email if of authenticated user.
    :param expires (str):       AWS Cognito session timeout.
    :param refresh_token (str): JWT refresh token received in respose.
    """
    session['username'] = username
    session['id'] = id
    session['groups'] = groups
    session['email'] = email
    session['expires'] = expires
    session['refresh_token'] = refresh_token


def verify(token: str, access_token: str = None):
    """
    Verifies a JWT string's signature and validates reserved claims.
    Get the key id from the header, locate it in the cognito keys and verify
    the key
    :param token (str):         A signed JWS to be verified.
    :param access_token (str):  An access token string. If the "at_hash" claim
                                is included in the
    :return id_token (dict):    The dict representation of the claims set,
                                assuming the signature is valid and all
                                requested data validation passes.
    """
    header = jwt.get_unverified_header(token)
    key = [k for k in config.jwt_cognito_key if k["kid"] == header['kid']][0]
    id_token = jwt.decode(token,
                          key,
                          audience=config.client_id,
                          access_token=access_token)
    return id_token


def logout_handler(fn):
    """
    A decorator to logout from AWS Cognito and return to signout uri.
    Use this decorator on the cognito logout endpoint.
    This handle will not return to handle any respose rather redirect to
    signout uri.
    This decorator also clears the basic informations from Flask session.
    Basic informations are:
        * username
        * group (List of Cognito groups if any)
        * id
        * email
        * expires
        * refresh_token
    """
    @wraps(fn)
    def wrapper(*args, **kwargs):
        update_session(username=None,
                       id=None,
                       groups=None,
                       email=None,
                       expires=None,
                       refresh_token=None)
        logger.info(
            "AWS Cognito Login, redirecting to AWS Cognito for logout and terminating sessions")

        aws_cognito_logout = config.logout_uri

        # https://docs.aws.amazon.com/cognito/latest/developerguide/amazon-cognito-user-pools-using-tokens-with-identity-providers.html
        res = redirect(aws_cognito_logout)
        return res
    return wrapper
